<?php

use Zen\Payment\Api;
use Zen\Payment\CartData;
use Zen\Payment\Notification;
use Zen\Payment\Paywall;
use Zen\Payment\Util;

/**
 * Class WC_Gateway_Zen_Payment
 *
 * @property string id
 * @property string icon
 * @property bool has_fields
 * @property string method_title
 * @property string method_description
 * @property string title
 * @property string description
 * @property array form_fields
 */
class WC_Gateway_Zen_Payment extends WC_Payment_Gateway
{
    /**
     * @var string
     */
    const PAYMENT_METHOD = 'zen_payment';

    /**
     * @var string
     */
    const PLUGIN_NAME = 'WooCommerce Zen Payment';

    const PLATFORM_NAME = 'WooCommerce';

    /**
     * @var string
     */
    const VERSION = '1.0.13';

    /**
     * initialise gateway with custom settings
     */
    public function __construct()
    {

        $this->id = self::PAYMENT_METHOD;
        $this->icon = self::get_image_src('logo.svg');
        $this->has_fields = false;
        $this->method_title = __('Zen', 'zen_payment');
        $this->method_description = __('Pay for purchases using the most convenient online payment methods.', 'zen_payment');
        $this->supports = [
            'products',
            'refunds',
        ];
        // User
        $this->title = $this->get_option('zen_payment_payment_title')
            ?: __('Online payment - Fast Bank Transfers, cards, BLIK.', 'zen_payment');
        $this->description = $this->get_option('description');

        $this->init_form_fields();
        $this->init_settings();

        //Actions
        add_action('woocommerce_update_options_payment_gateways_' . $this->id, [
            $this,
            'process_admin_options',
        ]);

        add_action('woocommerce_receipt_' . $this->id, [
            $this,
            'receipt_page',
        ]);

        add_action('woocommerce_api_' . strtolower(get_class($this)), [
            $this,
            'zen_notification',
        ]);

        //Filters
        add_filter('woocommerce_available_payment_gateways', [
            $this,
            'check_zen_available_payment_gateways',
        ]);

        //Assets
        add_action('admin_print_styles', function () {
            wp_enqueue_style('zen-payment-admin-css', plugin_dir_url(__FILE__) . '../resources/css/zen_payment-admin.css', [],
                time());
        }, 30);
    }

    /**
     * @param string $file
     *
     * @return string
     */
    public function get_image_src($file)
    {
        return WOOCOMMERCE_ZEN_PAYMENT_PLUGIN_URL . 'resources/images/' . $file;
    }

    /**
     * Initialize form in admin panel.
     */
    public function init_form_fields()
    {
        $currencies = Util::getCurrencies();

        $this->form_fields = [
            'zen_payment_enabled' => [
                'title' => __('Enable / Disable', 'zen_payment'),
                'type' => 'checkbox',
                'label' => __('Enable plugin', 'zen_payment'),
                'default' => 'yes',
            ],
            'zen_payment_terminal_id' => [
                'title' => __('Terminal UUID', 'zen_payment'),
                'type' => 'text',
                'default' => '',
            ],
            'zen_payment_paywall_secret' => [
                'title' => __('Paywall secret', 'zen_payment'),
                'type' => 'text',
                'default' => '',
            ],
            'zen_payment_ipn_secret' => [
                'title' => __('IPN secret', 'zen_payment'),
                'type' => 'text',
                'default' => '',
            ],
            'zen_payment_api_key' => [
                'title' => __('API key', 'zen_payment'),
                'type' => 'text',
                'default' => '',
            ],
            'zen_payment_payment_title' => [
                'title' => __('Payment title', 'zen_payment'),
                'type' => 'text',
                'placeholder' => $this->title,
            ],
            'zen_payment_currencies' => [
                'title' => __('Currency', 'woocommerce'),
                'type' => 'multiselect',
                'class' => 'wc-enhanced-select',
                'default' => 'taxable',
                'options' => $currencies,
            ],
        ];
    }

    /**
     * @param int $order_id
     *
     * @return array
     */
    public function process_payment($order_id)
    {

        $order = new WC_Order($order_id);

        //Reduce stock levels
	    wc_reduce_stock_levels($order_id);

        //Return thank you redirect
        return [
            'result' => 'success',
            'redirect' => $order->get_checkout_payment_url(true),
        ];
    }

    /**
     * @param string $order_id
     */
    public function receipt_page($order_id)
    {

        header("Cache-Control: no-cache, no-store, must-revalidate");

        $order = new WC_Order($order_id);
        $order_total = Util::convertAmount($order->get_total());

        $api = new Api(
            $this->get_option('zen_payment_api_key'),
            $this->get_environment()
        );

        $paymentData = Paywall::prepareOrderData(
            $order_total,
            $order->get_currency(),
            $order_id,
            $order->get_billing_first_name(),
            $order->get_billing_last_name(),
            $order->get_billing_email(),
            $this->get_return_url($order),
            $this->get_return_url($order),
            $this->get_return_url($order),
            add_query_arg('wc-api', 'wc_gateway_zen_payment', home_url('/')),
            $this->get_cart($order, $order_total),
            $this->get_option('zen_payment_terminal_id'),
            $this->get_option('zen_payment_paywall_secret'),
            self::PLUGIN_NAME,
            self::VERSION,
            self::PLATFORM_NAME,
            WC_VERSION
        );

        $paymentRequest = $api->createPayment(json_encode($paymentData));
        $paymentError = !$paymentRequest['success'];

	    // Remove cart
	    WC()->cart->empty_cart();

	    if (!$paymentError) {
		    wp_redirect($paymentRequest['body']['redirectUrl']);
	    } else {
		    wp_redirect($paymentRequest['data']['body']['error']['redirectUrl']);
	    }

	    exit();
    }

    /**
     * @return string
     */
    private function get_environment()
    {
        return Util::ENVIRONMENT_PRODUCTION;
    }

    /**
     * @param WC_Order $order
     * @param int $order_total
     *
     * @return array
     */
    public function get_cart($order, $order_total)
    {

        $wc = WC();

        $cart_data = new CartData();

        $cart_data->setAmount($order_total);

        foreach ($wc->cart->get_cart() as $cart_item) {

            $product = wc_get_product($cart_item['product_id']);
            $price = wc_get_price_including_tax($product);

            if ($product->is_type('variable')) {

                $product_variation = new WC_Product_Variation($cart_item['variation_id']);
                $price = wc_get_price_including_tax($product_variation);
            }

            $price = Util::convertAmount($price);

            $cart_data->addItem(
                $cart_item['data']->get_name(),
                $price,
                $cart_item['quantity'],
                Util::convertAmount($price * $cart_item['quantity'])
            );
        }

        $shipping_data = current($order->get_items('shipping'));

        if ($order->get_total_discount(false) > 0) {

            $discount = -abs(Util::convertAmount($order->get_total_discount(false)));

            $cart_data->addItem(
                __('Discount', 'zen_payment'),
                $discount,
                1,
                $discount
            );
        }

        if (!$shipping_data) {

            return $cart_data->getCartData(__('Round', 'zen_payment'));
        }

        $shipping_data = $shipping_data->get_data();

        foreach ($order->get_items('shipping') as $item_tax) {
            $shipping_data = json_decode($item_tax, true);
        }

        $shipping_cost = Util::convertAmount($shipping_data['total'] + $shipping_data['total_tax']);

        $cart_data->addItem(
            $shipping_data['name'],
            $shipping_cost,
            1,
            $shipping_cost);

        return $cart_data->getCartData(__('Round', 'zen_payment'));
    }

    /**
     * Method that checking notify and changing status
     *
     * @throws Exception
     */
    public function zen_notification()
    {

        $notification = new Notification(
            $this->get_option('zen_payment_ipn_secret')
        );

        // it can be order data or notification code - depends on verification notification
        $result_check_request_notification = $notification->checkRequest();

        if (is_int($result_check_request_notification)) {

            echo Notification::formatResponse(Notification::NS_ERROR);
            exit();
        }

        if ($result_check_request_notification['type'] === Notification::TRT_REFUND) {

            if ($result_check_request_notification['status'] !== Notification::STATUS_ACCEPTED) {
                echo Notification::formatResponse(Notification::NS_OK);
                exit();
            }

            if (!($order = wc_get_order(
                Util::extractMerchantTransactionId($result_check_request_notification['merchantTransactionId'])))
            ) {

                echo Notification::formatResponse(Notification::NS_ERROR);
                exit();
            }

            if ($order->get_status() === 'refunded') {
                echo Notification::formatResponse(Notification::NS_ERROR);
                exit();
            }

            $refund = wc_create_refund([
                'amount' => $result_check_request_notification['amount'],
                'reason' => 'Zen API',
                'order_id' => $order->get_id(),
                'refund_payment' => true,
            ]);

            if ($refund->errors) {

                self::log($refund->errors['error'][0]);

                echo Notification::formatResponse(Notification::NS_ERROR);
                exit();
            }

            $order->add_order_note(__('Refund for amount:', 'zen_payment')
                . ' '
                . $result_check_request_notification['amount']
                . ' '
                . __('with Zen UUID:', 'zen_payment')
                . ' '
                . $result_check_request_notification['transactionId']
                . ' '
                . __('has been correctly processed', 'zen_payment')
            );

            echo Notification::formatResponse(Notification::NS_OK);
            exit;
        }

        if (!($order = wc_get_order(Util::extractMerchantTransactionId($result_check_request_notification['merchantTransactionId'])))) {

            echo Notification::formatResponse(Notification::NS_ERROR);
            exit();
        }

        if (!Notification::checkRequestAmount($result_check_request_notification, $order->data['total'], $order->data['currency'])) {

            echo Notification::formatResponse(Notification::NS_ERROR);
            exit();
        }

        if ($order->get_status() === 'completed' || $order->get_status() === 'processing') {

            echo Notification::formatResponse(Notification::NS_OK);
            exit();
        }

        switch ($result_check_request_notification['status']) {
            case Notification::STATUS_ACCEPTED:
                $order->update_status($order->needs_processing() ? 'processing' : 'completed');
                $order->update_meta_data('zen_transaction_uuid', $result_check_request_notification['transactionId']);
                $order->save_meta_data();
                $order->add_order_note(__('Transaction reference', 'zen_payment') . ': ' . $result_check_request_notification['transactionId']);
                echo Notification::formatResponse(Notification::NS_OK);
                exit;
            case Notification::STATUS_REJECTED:
                $order->update_status('failed');
                $order->add_order_note(__('Transaction reference', 'zen_payment') . ': ' . $result_check_request_notification['transactionId']);
                echo Notification::formatResponse(Notification::NS_OK);
                exit;
            default:
                echo Notification::formatResponse(Notification::NS_OK);
                exit;
        }
    }

    /**
     * Logging method.
     *
     * @param string $message Log message.
     * @param string $level Optional. Default 'info'. Possible values:
     *                        emergency|alert|critical|error|warning|notice|info|debug.
     */
    public static function log($message, $level = 'info')
    {

        $log = wc_get_logger();
        $log->log($level, $message, ['source' => 'zen_payment']);

        return true;
    }

    /**
     * @param int $order_id
     * @param null $amount
     * @param string $reason
     *
     * @return bool
     * @throws Exception
     */
    public function process_refund($order_id, $amount = null, $reason = '')
    {

        if ($reason === 'Zen API') {
            return true;
        }

        $order = wc_get_order($order_id);

        $api = new Api(
            $this->get_option('zen_payment_api_key'),
            $this->get_environment()
        );

        $refund = $api->createRefund($amount, $order->get_meta('zen_transaction_uuid'), $order->get_currency(), $order_id);

        if ($refund['success']) {

            $order->add_order_note(
                __('Refund for amount:', 'zen_payment')
                . ' '
                . $amount
                . ' '
                . __('with Zen UUID:', 'zen_payment')
                . ' '
                . $refund['body']['id']
                . ' '
                . __('has been requested and waiting for completion.', 'zen_payment'));

            throw new Exception(__('A refund has been requested and waiting for completion.', 'zen_payment'));
        }

        return false;
    }

    /**
     * Validate data on checkout page. Remove Zen when there is an error.
     *
     * @param array $gateways
     *
     * @return array
     */
    public function check_zen_available_payment_gateways($gateways)
    {
        $currencies = $this->get_option('zen_payment_currencies');

        if (isset($currencies) && is_array($currencies)) {
            if (!Util::canUseForCurrency(get_woocommerce_currency(), $this->get_option('zen_payment_currencies'))
                || empty($this->get_option('zen_payment_terminal_id'))
                || empty($this->get_option('zen_payment_paywall_secret'))
                || empty($this->get_option('zen_payment_api_key'))
            ) {

                unset($gateways["zen_payment"]);
            }
        }

        return $gateways;
    }

}
